
/***************************************************************************
                          racecar.cpp  -  description
                             -------------------
    begin                : Wed Aug 21 01:49:10 EDT 2002
    copyright            : (C) 2002 by Kevin Greenhaw
    email                : kevin_greenhaw@email.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <qpointarray.h>
#include <qpen.h>

#include "racecar.h"
#include "LogTrace.h"

using namespace log_functions;

#ifndef M_PI
   #define M_PI 3.141592
#endif

#define toRadians(x) (x * (M_PI/180))
#define toDegrees(x) (x * (180/M_PI))
#define SQUARED(n) (n*n)

const double RaceCar::COS_15 = cos(toRadians(15));
const double RaceCar::COS_30 = cos(toRadians(30));
const double RaceCar::COS_45 = cos(toRadians(45));
const double RaceCar::COS_60 = cos(toRadians(60));
const double RaceCar::COS_75 = cos(toRadians(75));
const double RaceCar::SIN_15 = sin(toRadians(15));
const double RaceCar::SIN_30 = sin(toRadians(30));
const double RaceCar::SIN_45 = sin(toRadians(45));
const double RaceCar::SIN_60 = sin(toRadians(60));
const double RaceCar::SIN_75 = sin(toRadians(75));

//***************************************************************************//
//************************** CarLine Class **********************************//
//***************************************************************************//
/*****************************************************************************/
/* Function:     ImageItem                                                   */ 
/* Description:  Constructor                                                 */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
CarLine::CarLine( QCanvas *canvas)
    : QCanvasLine( canvas )
{
   QPen pen(QColor(0x00,0xFF,0xFF), 1);

   setZ(10);
   setPen(pen);
}

/*****************************************************************************/
/* Function:     SetImage                                                    */ 
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void CarLine::SetPoints(QRect &rect)
{
   int x1 = rect.left();
   int y1 = rect.top();
   int x2 = rect.right();
   int y2 = rect.bottom();

   setPoints(x1, y1, x2, y2);
}



//***************************************************************************//
//************************** RaceCar Class **********************************//
//***************************************************************************//
/*****************************************************************************/
/* Function:     RaceCar                                                     */ 
/* Description:  Constructor                                                 */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
RaceCar::RaceCar(QCanvasPixmapArray *s, QCanvas *c, CarControlType k) 
   : QCanvasSprite(s, c)
{ 
   CarControl           = k; 
   RaceCarDirection     = 0; 
   RaceCarSpeed         = 0; 
   LapsCompleted        = 0;
   RadiusAdjust         = 0;
   RaceCarPictureIndex  = 0;

   
   CollisionFlag        = 0;
   RestartFrameCounter  = 0;

   CarCollisionsIgnored = false;
   RaceCarRestarted     = false;
   TurnDelay            = true;
   
   TurnPoints.setAutoDelete(true);
   
   for (int i = 0; i < 16; i++)
   {
      TrackParts[i] = 0;
   }
}

/*****************************************************************************/
/* Function:     RaceCar                                                     */ 
/* Description:  Constructor                                                 */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::ResetRace()
{
   START_FUNCTION_TRACE("RaceCar::ResetRace");
   for (int i = 0; i < 16; i++)
   {
      TrackParts[i] = 0;
   }

   RaceCarRestarted     = false;
   RaceCarSpeedChange   = false;
   TurnDelay            = true;

   RaceCarDirection     = 0;
   RaceCarSpeed         = 0; 
   LapsCompleted        = 0;
   RadiusAdjust         = 0;
   
   CollisionFlag        = 0;
   RestartFrameCounter  = 0;
   RaceCarPictureIndex  = 0;
   END_FUNCTION_TRACE("RaceCar::ResetRace");
}

/*****************************************************************************/
/* Function:     SetRaceCarOffset                                            */
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::SetRaceCarOffset(int offsetX, int offsetY)
{
   START_FUNCTION_TRACE("RaceCar::SetRaceCarOffset");
   this->image()->setOffset(offsetX, offsetY);
   END_FUNCTION_TRACE("RaceCar::SetRaceCarOffset");
}

/*****************************************************************************/
/* Function:     SetRaceCarDirection                                         */
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::SetRaceCarSpeed(int Speed) 
{
   START_FUNCTION_TRACE("RaceCar::SetRaceCarSpeed");
   RaceCarSpeed = Speed;
   RaceCarSpeedChange = true;
   END_FUNCTION_TRACE("RaceCar::SetRaceCarSpeed");
}

/*****************************************************************************/
/* Function:     SetRaceCarDirection                                         */
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::SetRaceCarDirection(int Direction)
{
   START_FUNCTION_TRACE("RaceCar::SetRaceCarDirection");
   RaceCarDirection = Direction;

   if (CarControl == USER_CONTROL)
   {
      RaceCarPictureIndex = RaceCarDirection;
   }
   else
   {
      RaceCarPictureIndex = (RaceCarDirection == 0) ? 0  :
                           (RaceCarDirection == 2) ? 6  :
                           (RaceCarDirection == 4) ? 12 : 18;
   }
   END_FUNCTION_TRACE("RaceCar::SetRaceCarDirection");
}


/*****************************************************************************/
/* Function:     CheckForCollision                                           */
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
int RaceCar::CheckForCollision()
{
   START_FUNCTION_TRACE("RaceCar::CheckForCollision");
   QCanvasItemList list;
   if (CarControl == USER_CONTROL)
   {
      if (CarCollisionsIgnored == false)
      {
         list = QCanvasItem::collisions(true);
         for (QCanvasItemList::Iterator it=list.begin(); it!=list.end(); ++it) 
         {
            if ( (*it)->rtti() == raceCarRTTI )
            {
	            RaceCar *item= (RaceCar*)(*it);
               if (item->z() == this->z())
               {
                  if (item->GetCarControl() == COMPUTER_CONTROL)
                  {
                     CollisionFlag += 1;
                     setVelocity(0, 0);
                     RaceCarSpeed = 0;
                  }
               }
            }
         }
      }
   }

   END_FUNCTION_TRACE("RaceCar::CheckForCollision");
   return CollisionFlag;
}

/*****************************************************************************/
/* Function:     GetCarLocation                                           */ 
/* Description:                                                              */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::GetCarLocation(QRect *CarLocation)
{
   START_FUNCTION_TRACE("RaceCar::GetCarLocation");
   QRect thisCar = this->boundingRect();

   CarLocation->setTop(thisCar.top());
   CarLocation->setBottom(thisCar.bottom());
   CarLocation->setLeft(thisCar.left());
   CarLocation->setRight(thisCar.right());
   END_FUNCTION_TRACE("RaceCar::GetCarLocation");
}

/*****************************************************************************/
/* Function:     GetTurnDimensions                                           */ 
/* Description:  Calculate the Turn Radius and the Arc Length                */
/* Limitations:  Magic Offset needed to adjust for sloppy algorithm for each */
/*               track and each speed.                                       */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::GetTurnDimensions(int SideA, int SideB, float *Radius, int *ArcLength)
{
   START_FUNCTION_TRACE("RaceCar::GetTurnDimensions");
   float theta = (float)(M_PI / 2);
   
   *Radius = sqrt((SideA*SideA) + (SideB*SideB));
   

   // Necessary to adjust Radius Length to compensate for sloppy algorithm
   if (*Radius > 50)
   {
      *Radius = *Radius - (RadiusAdjust + 1);
   }
   else if (*Radius < 30)
   {
      *Radius = *Radius - (RadiusAdjust - 1);
   }
   else
   {
      *Radius = *Radius - RadiusAdjust;
   }
 
   *ArcLength = (int)((*Radius * theta) / RaceCarSpeed);
   END_FUNCTION_TRACE("RaceCar::GetTurnDimensions");
}

/*****************************************************************************/
/* Function:     CreateTurnPointArray                                        */
/* Description:  Calculate turn points and store them in an array.  This way */
/*               we only have to call the function once at the beginning of  */
/*               turn instead of at each update.                             */
/* Limitations:  Only works for turn radius of 90 degrees.                   */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::CreateTurnPointArray2(int       CurrentX, 
                                   int       CurrentY, 
                                   int       CarDirection,
                                   direction TrackDirection)
{
   START_FUNCTION_TRACE("RaceCar::CreateTurnPointArray2");
   float m, b, length, stepSize, steps;
   float newX, newY;
   int x1, x2, y1, y2;
   int sideA, sideB;

   const int TRACK_GROOVE = 40;
   // Find Slope
   x1 = CurrentX;
   y1 = CurrentY;
   x2 = 0;
   y2 = 0;

   switch( CarDirection )
   {
      case 1:
         if ( (TrackDirection == UP) || (TrackDirection == RIGHT) )
         {
            x2 = GRID_X;
//            y2 = CurrentX;

            y2 = TRACK_GROOVE;
         }
         else
         {
            x2 = TRACK_GROOVE;
//            x2 = CurrentY;
            y2 = 0;
         }

         break;

      case 3:
         if ( (TrackDirection == UP) || (TrackDirection == RIGHT) )
         {
            x2 = GRID_X - TRACK_GROOVE;
//            x2 = GRID_X - CurrentY;
            y2 = GRID_Y;
         }
         else
         {
            x2 = GRID_X;
//            y2 = GRID_Y - CurrentX;
            y2 = GRID_Y - TRACK_GROOVE;
         }

         break;

      case 5:
         if ( (TrackDirection == UP) || (TrackDirection == RIGHT) )
         {
            x2 = 0;
//            y2 = CurrentX;
            y2 = GRID_Y - TRACK_GROOVE;
         }
         else
         {
            x2 = GRID_X - TRACK_GROOVE;
//            x2 = CurrentY;
            y2 = GRID_Y;
         }

         break;

      case 7:
         if ( (TrackDirection == UP) || (TrackDirection == RIGHT) )
         {
            x2 = TRACK_GROOVE;
//            x2 = GRID_X - CurrentY;
            y2 = 0;
         }
         else
         {
            x2 = 0;
//            y2 = GRID_Y - CurrentX;
            y2 = TRACK_GROOVE;
         }

         break;
   }

   sideA = abs((y2 - y1));
   sideB = abs((x2 - x1));

   m = ((float)(y2 - y1)/(float)(x2 - x1));

   // Find Intercept
   b = CurrentY - (m * CurrentX);

   // Find Length
   length = sqrt((sideA*sideA) + (sideB*sideB)) / RaceCarSpeed;

   stepSize = sideB / length;

   TurnPoints.clear();
   ArcLength = 0;

   steps = 0;
   newX = CurrentX;

   while (steps < (sideB - stepSize))
   {
      if (CarDirection == 1)
      {
         // Calculate new point by using line equation y = mx + b
         newX += stepSize;
         newY = (m * newX) + b;
      }
      else if (CarDirection == 3)
      {
         // Calculate new point by using line equation y = mx + b
         newX += stepSize;
         newY = (m * newX) + b;
      }
      else if (CarDirection == 5)
      {
         // Calculate new point by using line equation y = mx + b
         newX -= stepSize;
         newY = (m * newX) + b;
      }
      else
      {
         // Calculate new point by using line equation y = mx + b
         newX -= stepSize;
         newY = (m * newX) + b;
      }


      TurnPoints.append(new QPoint((int)newX, (int)newY));

      steps += stepSize;
      ArcLength++;
   }

   TurnPoints.append(new QPoint(x2, y2));
   END_FUNCTION_TRACE("RaceCar::CreateTurnPointArray2");
}

/*****************************************************************************/
/* Function:     CreateTurnPointArray                                        */
/* Description:  Calculate turn points and store them in an array.  This way */
/*               we only have to call the function once at the beginning of  */
/*               turn instead of at each update.                             */
/* Limitations:  Only works for turn radius of 90 degrees.                   */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::CreateTurnPointArray(int       CurrentX, 
                                   int       CurrentY,
                                   int       CarDirection,
                                   direction TrackDirection)
{
   START_FUNCTION_TRACE("RaceCar::CreateTurnPointArray");

   double    theta,
             stepSize,
             arcSize;

   int       x1             = CurrentX, 
             y1             = CurrentY, 
             x2             = GRID_X, 
             y2             = GRID_Y,
             offset         = 10,
             sideA, 
             sideB,
             newX, 
             newY;
   
   TurnPoints.clear();
   ArcLength = 0;
   
   // Make race car slow down when going into a turn.  This makes the game 
   // a little more easy to win when in Medium and Hard levels of play.
   RaceCarSpeed = RaceCarSpeed - 1;
   
   switch (CarDirection)
   {
      case 1:
         if (TrackDirection == RIGHT)
         {
            sideA = (y2 - (GRID_Y/2) - offset);
            sideB = abs((x2 - x1));

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(0));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 1 TrackDirection RIGHT
               newX = static_cast<int>( GRID_X - (sideB * cos(theta)) );
               newY = static_cast<int>( GRID_Y - (sideA * sin(theta)) );

               theta += toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         else
         {
            sideA = y1;
            sideB = (x2 - (GRID_X/2) - offset);

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(90));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 1 TrackDirection LEFT -- (Y Offset Backwards)
               newX = static_cast<int>( (sideB * cos(theta)) );
               newY = static_cast<int>( (sideA * sin(theta)) );

               theta -= toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         break;

      case 3:
         if (TrackDirection == RIGHT)
         {
            sideA = abs((y2 - y1));
            sideB = (x2 - (GRID_X/2) - offset);

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(90));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 3 TrackDirection RIGHT
               newX = static_cast<int>( (sideB * cos(theta)) );
               newY = static_cast<int>( GRID_Y - (sideA * sin(theta)) );

               theta -= toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         else
         {
            sideA = (y2 - (GRID_Y/2) - offset);
            sideB = abs((x2 - x1));

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(0));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection3 TrackDirection LEFT
               newX = static_cast<int>( GRID_X  - (sideB * cos(theta)) );
               newY = static_cast<int>( (sideA * sin(theta)) );

               theta += toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         break;

      case 5:
         if (TrackDirection == RIGHT)
         {
            sideA = (y2 - (GRID_Y/2) - offset);
            sideB = x1;

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(0));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 5 TrackDirection RIGHT -- (X Offset Backwards)
               newX = static_cast<int>( (sideB * cos(theta)) );
               newY = static_cast<int>( (sideA * sin(theta)) );

               theta += toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         else
         {
            sideA = abs((y2 - y1));
            sideB = (x2 - (GRID_X/2) - offset);

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(90));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 5 TrackDirection LEFT
               newX = static_cast<int>( GRID_X  - (sideB * cos(theta)) );
               newY = static_cast<int>( GRID_Y - (sideA * sin(theta)) );

               theta -= toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         break;

      case 7:
         if (TrackDirection == RIGHT)
         {
            sideA = y1;
            sideB = (x2 - (GRID_X/2) - offset);

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(90));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection7 TrackDirection RIGHT -- (Y Offset Backwards)
               newX = static_cast<int>( GRID_X  - (sideB * cos(theta)) );
               newY = static_cast<int>( (sideA * sin(theta)) );

               theta -= toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         else
         {
            sideA = (y2 - (GRID_Y/2) - offset);
            sideB = x1;

            // ArcLength is equal to 1/4 Circumferenece of Circle using average
            // radius length in the equation.
            arcSize = ((sideA + sideB) / 2) * (0.5 * M_PI);
            stepSize  = toDegrees(0.5 * M_PI) / (arcSize/RaceCarSpeed);

            theta = static_cast<double>(toRadians(0));

            while ( (theta >= 0) && (theta <= toRadians(90)) )
            {
               // CarDirection 7 TrackDirection LEFT -- (X Offset Backwards)
               newX = static_cast<int>( (sideB * cos(theta)) );
               newY = static_cast<int>( GRID_Y - (sideA * sin(theta)) );

               theta += toRadians(stepSize);
               TurnPoints.append( new QPoint(newX, newY) );
               ArcLength++;
            }
         }
         break;
   }
   
   // Return race car to it's original speed.  The car will still go slower
   // through the turn because of the way the turn points are calculated.
   RaceCarSpeed = RaceCarSpeed + 1;

   END_FUNCTION_TRACE("RaceCar::CreateTurnPointArray");
}


/*****************************************************************************/
/* Function:     GetTurnIncrement                                            */
/* Description:  Calculate the Delta X and Delta Y for the next point in the */
/*               turn.                                                       */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::GetTurnIncrement(float *DeltaX, float *DeltaY)
{
   START_FUNCTION_TRACE("RaceCar::GetTurnIncrement");

   QPoint *p1;
   QPoint *p2;
   if ((RaceCarTurnRadius > 0) && (TurnPoints.at(RaceCarTurnRadius) != NULL))
   {
      p1 = TurnPoints.at(RaceCarTurnRadius-1);
      p2 = TurnPoints.at(RaceCarTurnRadius);

      *DeltaX = p2->x() - p1->x();
      *DeltaY = p2->y() - p1->y();
   }
   END_FUNCTION_TRACE("RaceCar::GetTurnIncrement");
}


/*****************************************************************************/
/* Function:     UpdateRaceCarDirection                                      */
/* Description:  Update the Race Car Picture index based on the location in  */
/*               the Turn.                                                   */
/* Limitations:  None                                                        */
/* Return Value: 0xFF if no update needed.                                   */
/*****************************************************************************/
int RaceCar::UpdateRaceCarDirection(int TurnRadius, direction TrackDirection,
                                    int MinPixIndex, int MaxPixIndex)
{
   START_FUNCTION_TRACE("RaceCar::UpdateRaceCarDirection");
   int PixIndex = 0xFF;

   if ( TurnRadius >= ((ArcLength / 5) * 4) )
   {
      if (TrackDirection == RIGHT)
      {
         PixIndex = MaxPixIndex;
      }
      else
      {
         PixIndex = MinPixIndex;
      }
   }
   else if ( TurnRadius >= ((ArcLength / 5) * 3) )
   {
      if (TrackDirection == RIGHT)
      {
         PixIndex = MaxPixIndex-1;
      }
      else
      {
         PixIndex = MinPixIndex+1;
      }
   }
   else if ( TurnRadius >= ((ArcLength / 5) * 2) )
   {
      if (TrackDirection == RIGHT)
      {
         PixIndex = MaxPixIndex-2;
      }
      else
      {
         PixIndex = MinPixIndex+2;
      }
   }
   else if ( TurnRadius >= (ArcLength / 5) )
   {
      if (TrackDirection == RIGHT)
      {
         PixIndex = MinPixIndex;
      }
      else
      {
         PixIndex = MaxPixIndex;
      }
   }

   END_FUNCTION_TRACE("RaceCar::UpdateRaceCarDirection");
   return PixIndex;
}


/*****************************************************************************/
/* Function:     UpdateCarDirection                                          */
/* Description:  Update the Direction of the Computer Controlled Race Car.   */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
int RaceCar::GetNewCarDirection(int       CarDirection,
                                int       GridDirection,
                                direction TurnDirection)
{
   START_FUNCTION_TRACE("RaceCar::GetNewCarDirection");
   int returnValue = 0;
   switch (CarDirection)
   {
      case 0:
         switch (GridDirection)
         {
            case 1:
            case 5:
               returnValue = ( ((TurnDirection == UP)||(TurnDirection == RIGHT)) ? 0 : 4 );
               break;

            default:
               returnValue = ( (TurnDirection == RIGHT) ? 1 : 7 );
               break;
         }
         break;

      case 2:
         switch (GridDirection)
         {
            case 1:
            case 5:
               returnValue = ( (TurnDirection == UP) ? 1 : 3 );
               break;

            case 2:
            case 6:
               returnValue = ( (TurnDirection == RIGHT) ? 2 : 6 );
               break;

            default:
               returnValue = ( (TurnDirection == RIGHT) ? 3 : 1 );
               break;
         }
         break;

      case 4:
         switch (GridDirection)
         {
            case 1:
            case 5:
               returnValue = ( ((TurnDirection == UP)||(TurnDirection == RIGHT)) ? 0 : 4 );
               break;

            default:
               returnValue = ( (TurnDirection == RIGHT) ? 5 : 3 );
               break;
         }
         break;

      case 6:
         switch (GridDirection)
         {
            case 1:
            case 5:
               returnValue = ( ((TurnDirection == UP)||(TurnDirection == RIGHT)) ? 7 : 5 );
               break;

            case 2:
            case 6:
               returnValue = ( (TurnDirection == RIGHT) ? 2 : 6 );
               break;

            default:
               returnValue = ( (TurnDirection == RIGHT) ? 7 : 5 );
               break;
         }         
         break;

      default:
         returnValue = 0;
         break;
   }
   
   END_FUNCTION_TRACE("RaceCar::GetNewCarDirection");
   return returnValue;
}


/*****************************************************************************/
/* Function:     UpdateCarDirection                                          */ 
/* Description:  Update the Direction of the Computer Controlled Race Car.   */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
int RaceCar::GetNewPictureIndex()
{
   START_FUNCTION_TRACE("RaceCar::GetNewPictureIndex");
   int returnValue = 0;
   
   switch (RaceCarDirection)
   {
      case 0:
         returnValue = 0;
         break;

      case 1:
         returnValue = ( PreviousCarDirection == 2 ) ? 5 : 1;
         break;

      case 2:
         returnValue = 6;
         break;

      case 3:
         returnValue = (PreviousCarDirection == 4) ? 11 : 7;
         break;

      case 4:
         returnValue = 12;
         break;

      case 5:
         returnValue = (PreviousCarDirection == 6) ? 17 : 13;
         break;

      case 6:
         returnValue = 18;
         break;

      case 7:
         returnValue = (PreviousCarDirection == 0) ? 23 : 19;
         break;
   }
   END_FUNCTION_TRACE("RaceCar::GetNewPictureIndex");
   return returnValue;
}


/*****************************************************************************/
/* Function:     UpdateCarDirection                                          */ 
/* Description:  Update the Direction of the Computer Controlled Race Car.   */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::SetVelocity(int CarDirection)
{
   START_FUNCTION_TRACE("RaceCar::SetVelocity");
   switch (CarDirection)
   {
      case 0:
         setVelocity(0, -RaceCarSpeed);
         break;

      case 2:
         setVelocity(RaceCarSpeed, 0);
         break;

      case 4:
         setVelocity(0, RaceCarSpeed);
         break;

      case 6:
         setVelocity(-RaceCarSpeed, 0);
         break;
   }
   END_FUNCTION_TRACE("RaceCar::SetVelocity");
}
 
/*****************************************************************************/
/* Function:     UpdateCarDirection                                          */ 
/* Description:  Update the Direction of the Computer Controlled Race Car.   */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
bool RaceCar::UpdateCarDirection(int       CarDirection,
                                 int       GridDirection,
                                 direction TrackDirection1,
                                 direction TrackDirection2,
                                 direction TrackDirection3)
{
   START_FUNCTION_TRACE("RaceCar::UpdateCarDirection");
   bool startTurn = false;

   PreviousCarDirection = CarDirection;
   SetVelocity(CarDirection);

   switch ( GridDirection )
   {
      case 1:     // Vertical Track
      case 5:
         TurnDirection = TrackDirection1;
         RaceCarDirection = GetNewCarDirection(CarDirection,
                                               GridDirection,
                                               TurnDirection);

         RaceCarPictureIndex = GetNewPictureIndex();

         if ( (CarDirection == 2) || (CarDirection == 6) )
         {
            startTurn         = true;
            TurnComplete      = false;
            RaceCarTurnRadius = 1;
         }
         else
         {
            TurnComplete = true;
         }
         break;

      case 2:     // Horizontal Track
      case 6:
         TurnDirection = TrackDirection2;
         RaceCarDirection = GetNewCarDirection(CarDirection,
                                               GridDirection,
                                               TurnDirection);

         RaceCarPictureIndex = GetNewPictureIndex();

         if ( (CarDirection == 0) || (CarDirection == 4) )
         {
            startTurn         = true;
            TurnComplete      = false;
            RaceCarTurnRadius = 1;
         }
         else
         {
            TurnComplete = true;
         }
         break;

      case 3:     // Curved Track
      case 4:
         TurnDirection = TrackDirection3;
         if ( (TurnComplete) || (!TurnDelay) )
         {

            RaceCarDirection = GetNewCarDirection(CarDirection,
                                                  GridDirection,
                                                  TurnDirection);

            RaceCarPictureIndex = GetNewPictureIndex();

            startTurn         = true;
            TurnComplete      = false;
            RaceCarTurnRadius = 1;
         }
         else
         {
            if (TurnDelay == false)
            {
               TurnComplete = true;
            }
         }
         break;

      default:
         break;
   }

   END_FUNCTION_TRACE("RaceCar::UpdateCarDirection");
   return startTurn;
}

/*****************************************************************************/
/* Function:     UpdateCarZLevel                                             */ 
/* Description:  Updates the Z Grid Level of the Race Car.                   */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::UpdateCarZLevel(int GridDirection)
{
   START_FUNCTION_TRACE("RaceCar::UpdateCarZLevel");
   if ( (GridDirection == 4) || (GridDirection == 5) || (GridDirection == 6) )
   {
      setZ(3);
   }
   else
   {
      setZ(1);
   }
   END_FUNCTION_TRACE("RaceCar::UpdateCarZLevel");
}

         
/*****************************************************************************/
/* Function:     UpdateComputerControlCar                                    */ 
/* Description:  Update the position of the Computer Controlled Race Car.    */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::UpdateComputerControlCar(int LocalX, int LocalY, GridSection *trackPiece)
{
   START_FUNCTION_TRACE("RaceCar::UpdateComputerControlCar");
   int  sideA = 0;
   int  sideB = 0;
   int  PixIndex;
   bool startTurn = false;

   static float newXDirection;
   static float newYDirection;

   static float radius = 0;

   UpdateCarZLevel(trackPiece->TrackGrid[LocalY][LocalX]);

   // Race Car Direction is not the Same as the Race Car Picture.  For a turn,
   // the race car direction is considered to be at a 45 degree angle in the
   // direction of the turn, while there are 5 different Race Car Pictures
   // that will be displayed during the turn.
   switch (GetRaceCarDirection())
   {
      case 0: // Race Car heading Up.
         startTurn = UpdateCarDirection(GetRaceCarDirection(),
                                        trackPiece->TrackGrid[LocalY][LocalX],
                                        trackPiece->GetTrackDirection(1),
                                        trackPiece->GetTrackDirection(2),
                                        trackPiece->GetTrackDirection(3));
         if (startTurn)
         {
            if (TurnDirection == RIGHT)
            {
               sideB = trackPiece->width() - LocalX;
            }
            else
            {
               sideB = LocalX;
            }

            sideA = trackPiece->height() - LocalY;
            GetTurnDimensions(sideA, sideB, &radius, &ArcLength);

            CreateTurnPointArray(LocalX, LocalY, RaceCarDirection, TurnDirection);
         }
         break;

      case 1:  // Race Car heading Up and to the Right
         if ( RaceCarTurnRadius == ArcLength )
         {
            switch ( trackPiece->TrackGrid[LocalY][LocalX] )
            {
               case 1:     // Vertical Track
               case 5:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(1) == UP) ? 0 : 4;
                  RaceCarPictureIndex = (RaceCarDirection == 0) ? 0 : 12;
                  break;

               case 2:     // Horizontal Track
               case 6:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(2) == RIGHT) ? 2 : 6;
                  RaceCarPictureIndex = (RaceCarDirection == 2) ? 6 : 18;
                  break;

               default:
                  RaceCarDirection = (PreviousCarDirection == 0) ? 2 : 0;
                  RaceCarPictureIndex = (RaceCarDirection == 2) ? 6 : 0;
            }
            SetVelocity(RaceCarDirection);
         }
         else        // Curved Track
         {
            TurnDelay = TRUE;
            GetTurnIncrement(&newXDirection, &newYDirection);

            setVelocity(newXDirection, newYDirection);

            RaceCarTurnRadius = RaceCarTurnRadius + 1;

            PixIndex = UpdateRaceCarDirection(RaceCarTurnRadius, TurnDirection,
                                                   2, 5);
            if (PixIndex != 0xFF)
            {
               RaceCarPictureIndex = PixIndex;
            }
         }
         break;

      case 2:     // Race Car heading to the Right
         startTurn = UpdateCarDirection(GetRaceCarDirection(),
                                        trackPiece->TrackGrid[LocalY][LocalX],
                                        trackPiece->GetTrackDirection(1),
                                        trackPiece->GetTrackDirection(2),
                                        trackPiece->GetTrackDirection(3));
         if (startTurn)
         {
            if (TurnDirection == UP)
            {
               sideA = LocalY;
            }
            else
            {
               sideA = trackPiece->height() - LocalY;
            }

            sideB = LocalX;
            GetTurnDimensions(sideA, sideB, &radius, &ArcLength);

            CreateTurnPointArray(LocalX, LocalY, RaceCarDirection, TurnDirection);
         }
         break;

      case 3:     // Race Car heading Down and to the Right
         if ( RaceCarTurnRadius == ArcLength )
         {
            switch ( trackPiece->TrackGrid[LocalY][LocalX] )
            {
               case 1:     // Vertical Track
               case 5:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(1) == UP) ? 0 : 4;
                  RaceCarPictureIndex = (RaceCarDirection == 0) ? 0 : 12;
                  break;

               case 2:     // Horizontal Track
               case 6:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(2) == RIGHT) ? 2 : 6;
                  RaceCarPictureIndex = (RaceCarDirection == 2) ? 6 : 18;
                  TurnComplete = true;
                  break;

               default:
                  RaceCarDirection = (PreviousCarDirection == 2) ? 4 : 2;
                  RaceCarPictureIndex = (RaceCarDirection == 4) ? 12 : 6;
            }
            SetVelocity(RaceCarDirection);
         }
         else
         {
            TurnDelay = TRUE;
            GetTurnIncrement(&newXDirection, &newYDirection);

            setVelocity(newXDirection, newYDirection);

            RaceCarTurnRadius = RaceCarTurnRadius + 1;

            PixIndex = UpdateRaceCarDirection(RaceCarTurnRadius, TurnDirection,
                                              8, 11);
            if (PixIndex != 0xFF)
            {
               RaceCarPictureIndex = PixIndex;
            }
         }
         break;

      case 4:     // Race Car heading Down
         startTurn = UpdateCarDirection(GetRaceCarDirection(),
                                        trackPiece->TrackGrid[LocalY][LocalX],
                                        trackPiece->GetTrackDirection(1),
                                        trackPiece->GetTrackDirection(2),
                                        trackPiece->GetTrackDirection(3));
         if (startTurn)
         {
            if (TurnDirection == RIGHT)
            {
               sideB = LocalX;
            }
            else
            {
               sideB = trackPiece->width() - LocalX;
            }

            sideA = LocalY;
            GetTurnDimensions(sideA, sideB, &radius, &ArcLength);

            CreateTurnPointArray(LocalX, LocalY, RaceCarDirection, TurnDirection);
         }
         break;

      case 5:     // Race Car heading Down and to the Left
         if ( RaceCarTurnRadius == ArcLength )
         {
            switch ( trackPiece->TrackGrid[LocalY][LocalX] )
            {
               case 1:     // Vertical Track
               case 5:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(1) == UP) ? 0 : 4;
                  RaceCarPictureIndex = (RaceCarDirection == 0) ? 0 : 12;
                  break;

               case 2:     // Horizontal Track
               case 6:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(2) == RIGHT) ? 2 : 6;
                  RaceCarPictureIndex = (RaceCarDirection == 2) ? 6 : 18;
                  break;

               default:
                  RaceCarDirection = (PreviousCarDirection == 4) ? 6 : 4;
                  RaceCarPictureIndex = (RaceCarDirection == 6) ? 18 : 12;

            }
            SetVelocity(RaceCarDirection);
         }
         else
         {
            TurnDelay = TRUE;
            GetTurnIncrement(&newXDirection, &newYDirection);

            setVelocity(newXDirection, newYDirection);

            RaceCarTurnRadius = RaceCarTurnRadius + 1;

            PixIndex = UpdateRaceCarDirection(RaceCarTurnRadius, TurnDirection,
                                              14, 17);
            if (PixIndex != 0xFF)
            {
               RaceCarPictureIndex = PixIndex;
            }
         }
         break;

      case 6:     // Race Car heading Left
         startTurn = UpdateCarDirection(GetRaceCarDirection(),
                                        trackPiece->TrackGrid[LocalY][LocalX],
                                        trackPiece->GetTrackDirection(1),
                                        trackPiece->GetTrackDirection(2),
                                        trackPiece->GetTrackDirection(3));
         if (startTurn)
         {
            if (TurnDirection == UP)
            {
               sideA = LocalY;
            }
            else
            {
               sideA = trackPiece->height() - LocalY;
            }

            sideB = trackPiece->width() - LocalX;
            GetTurnDimensions(sideA, sideB, &radius, &ArcLength);

            CreateTurnPointArray(LocalX, LocalY, RaceCarDirection, TurnDirection);
         }
         break;

      case 7:     // Race Car heading Up and to the Left
         if ( RaceCarTurnRadius == ArcLength )
         {
            switch ( trackPiece->TrackGrid[LocalY][LocalX] )
            {
               case 1:     // Vertical Track
               case 5:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(1) == UP) ? 0 : 4;
                  RaceCarPictureIndex = (RaceCarDirection == 0) ? 0 : 12;
                  break;

               case 2:     // Horizontal Track
               case 6:
                  RaceCarDirection =
                     (trackPiece->GetTrackDirection(2) == RIGHT) ? 2 : 6;
                  RaceCarPictureIndex = (RaceCarDirection == 2) ? 6 : 18;
                  break;

               default:
                  RaceCarDirection = (PreviousCarDirection == 6) ? 0 : 6;
                  RaceCarPictureIndex = (RaceCarDirection == 0) ? 0 : 18;

            }
            SetVelocity(RaceCarDirection);
         }
         else
         {
            TurnDelay = TRUE;
            GetTurnIncrement(&newXDirection, &newYDirection);

            setVelocity(newXDirection, newYDirection);

            RaceCarTurnRadius = RaceCarTurnRadius + 1;

            PixIndex = UpdateRaceCarDirection(RaceCarTurnRadius, TurnDirection,
                                              20, 23);
            if (PixIndex != 0xFF)
            {
               RaceCarPictureIndex = PixIndex;
            }
         }
         break;

      default:
         break;
   }
   END_FUNCTION_TRACE("RaceCar::UpdateComputerControlCar");
}

/*****************************************************************************/
/* Function:     UpdateUserControlCar                                        */
/* Description:  Update the position of the User Controlled Race Car.        */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
QRect RaceCar::GetRaceCarCenterline()
{
   START_FUNCTION_TRACE("RaceCar::GetRaceCarCenterline");
   QRect imageRect = this->boundingRect();
   QRect returnRect;

   int x1 = imageRect.left();
   int y1 = imageRect.top();
   int x2 = imageRect.right();
   int y2 = imageRect.bottom();

   switch (RaceCarPictureIndex)
   {
      case 0:
      case 12:
         returnRect.setCoords(x1+10, y1, x2-10, y2);
         break;
           
      case 1:
      case 13:
         returnRect.setCoords(x1+13, y1, x2-13, y2);
         break;
           
      case 2:
      case 14:
         returnRect.setCoords(x1+16, y1, x2-16, y2);
         break;
           
      case 3:
      case 15:
         returnRect.setCoords(x2, y1, x1, y2);
         break;
           
      case 4:
      case 16:
         returnRect.setCoords(x1, y1+16, x2, y2-16);
         break;
           
      case 5:
      case 17:
         returnRect.setCoords(x1, y1+13, x2, y2-13);
         break;
           
      case 6:
      case 18:
         returnRect.setCoords(x1, y1+10, x2, y2-10);
         break;
           
      case 7:
      case 19:
         returnRect.setCoords(x1, y1+7, x2, y2-7);
         break;
           
      case 8:
      case 20:
         returnRect.setCoords(x1, y1+4, x2, y2-4);
         break;
           
      case 9:
      case 21:
         returnRect.setCoords(x1, y1, x2, y2);
         break;
           
      case 10:
      case 22:
         returnRect.setCoords(x1+4, y1, x2-4, y2);
         break;
           
      case 11:
      case 23:
         returnRect.setCoords(x1+7, y1, x2-7, y2);
         break;

      default:
         returnRect.setCoords(0,0,0,0);
   }       

   END_FUNCTION_TRACE("RaceCar::GetRaceCarCenterline");
   return returnRect;
}


/*****************************************************************************/
/* Function:     UpdateUserControlCar                                        */
/* Description:  Update the position of the User Controlled Race Car.        */
/* Limitations:  None                                                        */
/* Return Value: None                                                        */
/*****************************************************************************/
void RaceCar::UpdateUserControlCar(int LocalX, int LocalY, GridSection *trackPiece)
{
   START_FUNCTION_TRACE("RaceCar::UpdateUserControlCar");
   static bool OutOfBounds = false;

   if (RaceCarSpeed > 0)
   {
      if ( (OutOfBounds == false) && (trackPiece->TrackGrid[LocalY][LocalX] != 0) )
      {
         if ( (trackPiece->TrackGrid[LocalY][LocalX] >= 4) &&
              (trackPiece->TrackGrid[LocalY][LocalX] <= 6) )
         {
            setZ(3);
         }
         else
         {
            setZ(1);
         }
         OutOfBounds = false;

         // For the User Controlled car, Race Car Direction is the same as the
         // Race Car Picture.
         switch (RaceCarDirection)
         {
            case 0:
               setVelocity(0, -RaceCarSpeed);
               break;
            
            case 1:
               setVelocity(COS_75 * RaceCarSpeed, 
                           -SIN_75 * RaceCarSpeed);
               break;
            
            case 2:
               setVelocity(COS_60 * RaceCarSpeed, 
                           -SIN_60 * RaceCarSpeed);
               break;
            
            case 3:
               setVelocity(COS_45 * RaceCarSpeed, 
                           -SIN_45 * RaceCarSpeed);
               break;
            
            case 4:
               setVelocity(COS_30 * RaceCarSpeed,
                           -SIN_30 * RaceCarSpeed);
               break;

            case 5:
               setVelocity(COS_15 * RaceCarSpeed,
                           -SIN_15 * RaceCarSpeed);
               break;

            case 6:
               setVelocity(RaceCarSpeed, 0);
               break;

            case 7:
               setVelocity(COS_15 * RaceCarSpeed,
                           SIN_15 * RaceCarSpeed);
               break;

            case 8:
               setVelocity(COS_30 * RaceCarSpeed,
                           SIN_30 * RaceCarSpeed);
               break;

            case 9:
               setVelocity(COS_45 * RaceCarSpeed,
                           SIN_45 * RaceCarSpeed);
               break;

            case 10:
               setVelocity(COS_60 * RaceCarSpeed,
                           SIN_60 * RaceCarSpeed);
               break;

            case 11:
               setVelocity(COS_75 * RaceCarSpeed,
                           SIN_75 * RaceCarSpeed);
               break;

            case 12:
               setVelocity(0, RaceCarSpeed);
               break;

            case 13:
               setVelocity(-COS_75 * RaceCarSpeed,
                           SIN_75 * RaceCarSpeed);
               break;

            case 14:
               setVelocity(-COS_60 * RaceCarSpeed,
                           SIN_60 * RaceCarSpeed);
               break;

            case 15:
               setVelocity(-COS_45 * RaceCarSpeed,
                           SIN_45 * RaceCarSpeed);
               break;

            case 16:
               setVelocity(-COS_30 * RaceCarSpeed,
                           SIN_30 * RaceCarSpeed);
               break;

            case 17:
               setVelocity(-COS_15 * RaceCarSpeed,
                           SIN_15 * RaceCarSpeed);
               break;

            case 18:
               setVelocity(-RaceCarSpeed, 0);
               break;

            case 19:
               setVelocity(-COS_15 * RaceCarSpeed,
                           -SIN_15 * RaceCarSpeed);
               break;

            case 20:
               setVelocity(-COS_30 * RaceCarSpeed,
                           -SIN_30 * RaceCarSpeed);
               break;

            case 21:
               setVelocity(-COS_45 * RaceCarSpeed,
                           -SIN_45 * RaceCarSpeed);
               break;

            case 22:
               setVelocity(-COS_60 * RaceCarSpeed,
                           -SIN_60 * RaceCarSpeed);
               break;

            case 23:
               setVelocity(-COS_75 * RaceCarSpeed,
                           -SIN_75 * RaceCarSpeed);
               break;
         }
      }
      else
      {
         CollisionFlag += 1;
      }
   }
   else
   {
      setVelocity(0,0);
   }
   
   // MAX_COLLISIONS is used to add some sloppiness into the collisions
   // instead of registering a collision as soon as the car touches the
   // edge of the track, this gives a better feel to the game.
   if (CollisionFlag > MAX_COLLISIONS)
   {
      setVelocity(0,0);
      RaceCarSpeed = 0;
   }

   RaceCarPictureIndex = RaceCarDirection;
   END_FUNCTION_TRACE("RaceCar::UpdateUserControlCar");
}
